from __future__ import annotations

from dataclasses import dataclass
from math import inf, nan
from typing import TYPE_CHECKING, ClassVar, Literal, cast

import redis
import redis.asyncio
from hypothesis import assume, given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    data,
    floats,
    sampled_from,
    tuples,
)
from polars import Boolean, DataFrame, DataType, Float64, Int64, Utf8
from polars.testing import assert_frame_equal
from pytest import mark, param, raises
from redis.commands.timeseries import TimeSeries

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from utilities.datetime import EPOCH_UTC, drop_microseconds
from utilities.hypothesis import (
    YieldRedisContainer,
    int32s,
    lists_fixed_length,
    redis_cms,
    redis_cms_async,
    text_ascii,
    zoned_datetimes,
)
from utilities.polars import DatetimeUTC, check_polars_dataframe, zoned_datetime
from utilities.redis import (
    TimeSeriesAddDataFrameError,
    TimeSeriesAddError,
    TimeSeriesMAddError,
    TimeSeriesRangeError,
    TimeSeriesReadDataFrameError,
    ensure_time_series_created,
    ensure_time_series_created_async,
    time_series_add,
    time_series_add_async,
    time_series_add_dataframe,
    time_series_add_dataframe_async,
    time_series_get,
    time_series_get_async,
    time_series_madd,
    time_series_madd_async,
    time_series_range,
    time_series_range_async,
    time_series_read_dataframe,
    time_series_read_dataframe_async,
    yield_client,
    yield_client_async,
    yield_time_series,
    yield_time_series_async,
)
from utilities.zoneinfo import HONG_KONG, UTC

if TYPE_CHECKING:
    import datetime as dt
    from zoneinfo import ZoneInfo

    from polars._typing import PolarsDataType, SchemaDict

    from utilities.types import Number

valid_zoned_datetimes = zoned_datetimes(
    min_value=EPOCH_UTC, time_zone=sampled_from([HONG_KONG, UTC]), valid=True
).map(drop_microseconds)
invalid_zoned_datetimes = (
    zoned_datetimes(
        max_value=EPOCH_UTC, time_zone=sampled_from([HONG_KONG, UTC]), valid=True
    )
    .map(drop_microseconds)
    .filter(lambda t: t < EPOCH_UTC)
)


@SKIPIF_CI_AND_NOT_LINUX
class TestEnsureTimeSeriesCreated:
    @given(yield_redis=redis_cms())
    def test_sync(self, *, yield_redis: YieldRedisContainer) -> None:
        with yield_redis() as redis:
            assert redis.client.exists(redis.key) == 0
            for _ in range(2):
                ensure_time_series_created(redis.client.ts(), redis.key)
            assert redis.client.exists(redis.key) == 1

    @given(data=data())
    async def test_async(self, *, data: DataObject) -> None:
        async with redis_cms_async(data) as redis:
            assert await redis.client.exists(redis.key) == 0
            for _ in range(2):
                await ensure_time_series_created_async(redis.client.ts(), redis.key)
            assert await redis.client.exists(redis.key) == 1


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndGet:
    @given(
        yield_redis=redis_cms(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_sync(
        self, *, yield_redis: YieldRedisContainer, timestamp: dt.datetime, value: float
    ) -> None:
        with yield_redis() as redis:
            result = time_series_add(redis.ts, redis.key, timestamp, value)
            assert isinstance(result, int)
            res_timestamp, res_value = time_series_get(redis.ts, redis.key)
            assert res_timestamp == timestamp.astimezone(UTC)
            assert res_value == value

    @given(
        yield_redis=redis_cms(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_sync_error_at_upsert(
        self, *, yield_redis: YieldRedisContainer, timestamp: dt.datetime, value: float
    ) -> None:
        with yield_redis() as redis:
            _ = time_series_add(redis.ts, redis.key, timestamp, value)
            with raises(
                TimeSeriesAddError,
                match="Error at upsert under DUPLICATE_POLICY == 'BLOCK'; got .*",
            ):
                _ = time_series_add(redis.ts, redis.key, timestamp, value)

    @given(
        yield_redis=redis_cms(),
        timestamp=invalid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_sync_invalid_timestamp(
        self, *, yield_redis: YieldRedisContainer, timestamp: dt.datetime, value: float
    ) -> None:
        _ = assume(timestamp < EPOCH_UTC)
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesAddError, match="Timestamp must be at least the Epoch; got .*"
            ),
        ):
            _ = time_series_add(redis.ts, redis.key, timestamp, value)

    @given(yield_redis=redis_cms(), timestamp=valid_zoned_datetimes)
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    def test_sync_invalid_value(
        self, *, yield_redis: YieldRedisContainer, timestamp: dt.datetime, value: float
    ) -> None:
        with (
            yield_redis() as redis,
            raises(TimeSeriesAddError, match="Invalid value; got .*"),
        ):
            _ = time_series_add(redis.ts, redis.key, timestamp, value)

    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_async(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        async with redis_cms_async(data) as redis:
            result = await time_series_add_async(redis.ts, redis.key, timestamp, value)
            assert isinstance(result, int)
            res_timestamp, res_value = await time_series_get_async(redis.ts, redis.key)
            assert res_timestamp == timestamp.astimezone(UTC)
            assert res_value == value

    @given(
        data=data(),
        timestamp=valid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_async_error_at_upsert(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        async with redis_cms_async(data) as redis:
            _ = await time_series_add_async(redis.ts, redis.key, timestamp, value)
            with raises(
                TimeSeriesAddError,
                match="Error at upsert under DUPLICATE_POLICY == 'BLOCK'; got .*",
            ):
                _ = await time_series_add_async(redis.ts, redis.key, timestamp, value)

    @given(
        data=data(),
        timestamp=invalid_zoned_datetimes,
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    async def test_async_invalid_timestamp(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        _ = assume(timestamp < EPOCH_UTC)
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesAddError, match="Timestamp must be at least the Epoch; got .*"
            ):
                _ = await time_series_add_async(redis.ts, redis.key, timestamp, value)

    @given(data=data(), timestamp=valid_zoned_datetimes)
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    async def test_async_invalid_value(
        self, *, data: DataObject, timestamp: dt.datetime, value: float
    ) -> None:
        async with redis_cms_async(data) as redis:
            with raises(TimeSeriesAddError, match="Invalid value; got .*"):
                _ = await time_series_add_async(redis.ts, redis.key, timestamp, value)


@dataclass(frozen=True, kw_only=True)
class _TestTimeSeriesAddAndReadDataFramePrepare:
    df: DataFrame
    key: str
    timestamp: str
    keys: tuple[str, str]
    columns: tuple[str, str]
    time_zone: ZoneInfo
    schema: SchemaDict


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndReadDataFrame:
    schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Float64,
    }

    @given(
        data=data(),
        yield_redis=redis_cms(),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
        key_timestamp_values=lists_fixed_length(text_ascii(), 4, unique=True).map(
            tuple
        ),
        time_zone=sampled_from([HONG_KONG, UTC]),
    )
    @mark.parametrize(
        ("strategy1", "dtype1"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    @mark.parametrize(
        ("strategy2", "dtype2"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    def test_sync(
        self,
        *,
        data: DataObject,
        yield_redis: YieldRedisContainer,
        series_names: tuple[str, str],
        strategy1: SearchStrategy[Number],
        strategy2: SearchStrategy[Number],
        key_timestamp_values: tuple[str, str, str, str],
        time_zone: ZoneInfo,
        dtype1: DataType,
        dtype2: DataType,
    ) -> None:
        with yield_redis() as redis:
            prepared = self._prepare_main_test(
                data,
                redis.key,
                series_names,
                strategy1,
                strategy2,
                key_timestamp_values,
                time_zone,
                dtype1,
                dtype2,
            )
            time_series_add_dataframe(
                redis.ts, prepared.df, key=prepared.key, timestamp=prepared.timestamp
            )
            result = time_series_read_dataframe(
                redis.ts,
                prepared.keys,
                prepared.columns,
                output_key=prepared.key,
                output_timestamp=prepared.timestamp,
                output_time_zone=prepared.time_zone,
            )
            check_polars_dataframe(result, height=2, schema_list=prepared.schema)
        assert_frame_equal(result, prepared.df)

    @given(yield_redis=redis_cms())
    def test_sync_error_add_key_missing(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame()
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesAddDataFrameError,
                match="DataFrame must have a 'key' column; got .*",
            ),
        ):
            _ = time_series_add_dataframe(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_add_timestamp_missing(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesAddDataFrameError,
                match="DataFrame must have a 'timestamp' column; got .*",
            ),
        ):
            _ = time_series_add_dataframe(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_add_key_is_not_utf8(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Boolean, "timestamp": DatetimeUTC})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesAddDataFrameError,
                match="The 'key' column must be Utf8; got Boolean",
            ),
        ):
            _ = time_series_add_dataframe(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesAddDataFrameError,
                match="The 'timestamp' column must be a zoned Datetime; got Boolean",
            ),
        ):
            _ = time_series_add_dataframe(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_read_no_keys_requested(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesReadDataFrameError, match="At least 1 key must be requested"
            ),
        ):
            _ = time_series_read_dataframe(redis.ts, [], [])

    @given(yield_redis=redis_cms())
    def test_sync_error_read_no_columns_requested(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesReadDataFrameError,
                match="At least 1 column must be requested",
            ),
        ):
            _ = time_series_read_dataframe(redis.ts, redis.key, [])

    @given(
        data=data(),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
        key_timestamp_values=lists_fixed_length(text_ascii(), 4, unique=True).map(
            tuple
        ),
        time_zone=sampled_from([HONG_KONG, UTC]),
    )
    @mark.parametrize(
        ("strategy1", "dtype1"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    @mark.parametrize(
        ("strategy2", "dtype2"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    async def test_async(
        self,
        *,
        data: DataObject,
        series_names: tuple[str, str],
        strategy1: SearchStrategy[Number],
        strategy2: SearchStrategy[Number],
        key_timestamp_values: tuple[str, str, str, str],
        time_zone: ZoneInfo,
        dtype1: DataType,
        dtype2: DataType,
    ) -> None:
        async with redis_cms_async(data) as redis:
            prepared = self._prepare_main_test(
                data,
                redis.key,
                series_names,
                strategy1,
                strategy2,
                key_timestamp_values,
                time_zone,
                dtype1,
                dtype2,
            )
            await time_series_add_dataframe_async(
                redis.ts, prepared.df, key=prepared.key, timestamp=prepared.timestamp
            )
            result = await time_series_read_dataframe_async(
                redis.ts,
                prepared.keys,
                prepared.columns,
                output_key=prepared.key,
                output_timestamp=prepared.timestamp,
                output_time_zone=prepared.time_zone,
            )
            check_polars_dataframe(result, height=2, schema_list=prepared.schema)
        assert_frame_equal(result, prepared.df)

    @given(data=data())
    async def test_async_error_add_key_missing(self, *, data: DataObject) -> None:
        df = DataFrame()
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesAddDataFrameError,
                match="DataFrame must have a 'key' column; got .*",
            ):
                _ = await time_series_add_dataframe_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_add_timestamp_missing(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesAddDataFrameError,
                match="DataFrame must have a 'timestamp' column; got .*",
            ):
                _ = await time_series_add_dataframe_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_add_key_is_not_utf8(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Boolean, "timestamp": DatetimeUTC})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesAddDataFrameError,
                match="The 'key' column must be Utf8; got Boolean",
            ):
                _ = await time_series_add_dataframe_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesAddDataFrameError,
                match="The 'timestamp' column must be a zoned Datetime; got Boolean",
            ):
                _ = await time_series_add_dataframe_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_read_no_keys_requested(
        self, *, data: DataObject
    ) -> None:
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesReadDataFrameError, match="At least 1 key must be requested"
            ):
                _ = await time_series_read_dataframe_async(redis.ts, [], [])

    @given(data=data())
    async def test_async_error_read_no_columns_requested(
        self, *, data: DataObject
    ) -> None:
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesReadDataFrameError,
                match="At least 1 column must be requested",
            ):
                _ = await time_series_read_dataframe_async(redis.ts, redis.key, [])

    def _prepare_main_test(
        self,
        data: DataObject,
        redis_key: str,
        series_names: tuple[str, str],
        strategy1: SearchStrategy[Number],
        strategy2: SearchStrategy[Number],
        key_timestamp_values: tuple[str, str, str, str],
        time_zone: ZoneInfo,
        dtype1: DataType,
        dtype2: DataType,
        /,
    ) -> _TestTimeSeriesAddAndReadDataFramePrepare:
        key1, key2 = keys = cast(
            tuple[str, str], tuple(f"{redis_key}_{id_}" for id_ in series_names)
        )
        timestamp1, timestamp2 = data.draw(
            tuples(valid_zoned_datetimes, valid_zoned_datetimes)
        )
        value11, value21 = data.draw(tuples(strategy1, strategy1))
        value12, value22 = data.draw(tuples(strategy2, strategy2))
        key, timestamp, column1, column2 = key_timestamp_values
        schema = {
            key: Utf8,
            timestamp: zoned_datetime(time_zone=time_zone),
            column1: dtype1,
            column2: dtype2,
        }
        df = DataFrame(
            [
                (key1, timestamp1, value11, value12),
                (key2, timestamp2, value21, value22),
            ],
            schema=schema,
            orient="row",
        )
        return _TestTimeSeriesAddAndReadDataFramePrepare(
            df=df,
            key=key,
            timestamp=timestamp,
            keys=keys,
            columns=(column1, column2),
            time_zone=time_zone,
            schema=schema,
        )


@dataclass(frozen=True, kw_only=True)
class _TestTimeSeriesMAddAndRangePrepare:
    keys: tuple[str, str]
    triples: list[tuple[str, dt.datetime, Number]]
    key: str
    timestamp: str
    value: str
    values_or_df: list[tuple[str, dt.datetime, Number]] | DataFrame
    schema: SchemaDict


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesMAddAndRange:
    int_schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Int64,
    }
    float_schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Float64,
    }

    @given(
        data=data(),
        yield_redis=redis_cms(),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
        time_zone=sampled_from([HONG_KONG, UTC]),
        key_timestamp_value=lists_fixed_length(text_ascii(), 3, unique=True).map(tuple),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize(
        ("strategy", "dtype"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    def test_sync(
        self,
        *,
        data: DataObject,
        yield_redis: YieldRedisContainer,
        series_names: tuple[str, str],
        time_zone: ZoneInfo,
        key_timestamp_value: tuple[str, str, str],
        case: Literal["values", "DataFrame"],
        strategy: SearchStrategy[Number],
        dtype: PolarsDataType,
    ) -> None:
        with yield_redis() as redis:
            prepared = self._prepare_main_test(
                data,
                redis.key,
                series_names,
                time_zone,
                key_timestamp_value,
                case,
                strategy,
                dtype,
            )
            res_madd = time_series_madd(
                redis.ts,
                prepared.values_or_df,
                key=prepared.key,
                timestamp=prepared.timestamp,
                value=prepared.value,
            )
            assert isinstance(res_madd, list)
            for i in res_madd:
                assert isinstance(i, int)
            res_range = time_series_range(
                redis.ts,
                prepared.keys,
                output_key=prepared.key,
                output_timestamp=prepared.timestamp,
                output_time_zone=time_zone,
                output_value=prepared.value,
            )
            check_polars_dataframe(res_range, height=2, schema_list=prepared.schema)
            assert res_range.rows() == prepared.triples

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_key_missing(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame()
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError, match="DataFrame must have a 'key' column; got .*"
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_timestamp_missing(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError,
                match="DataFrame must have a 'timestamp' column; got .*",
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_value_missing(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError,
                match="DataFrame must have a 'value' column; got .*",
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_key_is_not_utf8(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(
            schema={"key": Boolean, "timestamp": DatetimeUTC, "value": Float64}
        )
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError, match="The 'key' column must be Utf8; got Boolean"
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean, "value": Float64})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError,
                match="The 'timestamp' column must be a zoned Datetime; got Boolean",
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(yield_redis=redis_cms())
    def test_sync_error_madd_value_is_not_numeric(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC, "value": Boolean})
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesMAddError,
                match="The 'value' column must be numeric; got Boolean",
            ),
        ):
            _ = time_series_madd(redis.ts, df)

    @given(data=data(), yield_redis=redis_cms())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    def test_sync_error_madd_invalid_key(
        self,
        *,
        data: DataObject,
        yield_redis: YieldRedisContainer,
        case: Literal["values", "DataFrame"],
    ) -> None:
        with yield_redis() as redis:
            values_or_df = self._prepare_test_error_madd_invalid_key(
                data, redis.key, case
            )
            with raises(TimeSeriesMAddError, match="The key '.*' must exist"):
                _ = time_series_madd(
                    redis.ts, values_or_df, assume_time_series_exist=True
                )

    @given(data=data(), yield_redis=redis_cms())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    def test_sync_error_madd_invalid_timestamp(
        self,
        *,
        data: DataObject,
        yield_redis: YieldRedisContainer,
        case: Literal["values", "DataFrame"],
    ) -> None:
        with yield_redis() as redis:
            values_or_df = self._prepare_test_error_madd_invalid_timestamp(
                data, redis.key, case
            )
            with raises(
                TimeSeriesMAddError,
                match="Timestamps must be at least the Epoch; got .*",
            ):
                _ = time_series_madd(redis.ts, values_or_df)

    @given(data=data(), yield_redis=redis_cms())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    def test_sync_error_madd_invalid_value(
        self,
        *,
        data: DataObject,
        yield_redis: YieldRedisContainer,
        case: Literal["values", "DataFrame"],
        value: float,
    ) -> None:
        with yield_redis() as redis:
            values_or_df = self._prepare_test_error_madd_invalid_value(
                data, redis.key, case, value
            )
            with raises(TimeSeriesMAddError, match="The value .* is invalid"):
                _ = time_series_madd(redis.ts, values_or_df)

    @given(yield_redis=redis_cms())
    def test_sync_error_range_no_keys_requested(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        with (
            yield_redis() as redis,
            raises(
                TimeSeriesRangeError, match="At least 1 key must be requested; got .*"
            ),
        ):
            _ = time_series_range(redis.ts, [])

    @given(yield_redis=redis_cms())
    def test_sync_error_range_invalid_key(
        self, *, yield_redis: YieldRedisContainer
    ) -> None:
        with (
            yield_redis() as redis,
            raises(TimeSeriesRangeError, match="The key '.*' must exist"),
        ):
            _ = time_series_range(redis.ts, redis.key)

    @given(data=data(), yield_redis=redis_cms())
    def test_sync_error_range_key_with_int64_and_float64(
        self, *, data: DataObject, yield_redis: YieldRedisContainer
    ) -> None:
        with yield_redis() as redis:
            values = self._prepare_test_error_range_key_with_int64_and_float64(
                data, redis.key
            )
            for vals in values:
                _ = time_series_madd(redis.ts, vals)
            with raises(
                TimeSeriesRangeError,
                match="The key '.*' contains both Int64 and Float64 data",
            ):
                _ = time_series_range(redis.ts, redis.key)

    @given(
        data=data(),
        series_names=lists_fixed_length(text_ascii(), 2, unique=True).map(tuple),
        time_zone=sampled_from([HONG_KONG, UTC]),
        key_timestamp_value=lists_fixed_length(text_ascii(), 3, unique=True).map(tuple),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize(
        ("strategy", "dtype"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    async def test_async(
        self,
        *,
        data: DataObject,
        series_names: tuple[str, str],
        time_zone: ZoneInfo,
        key_timestamp_value: tuple[str, str, str],
        case: Literal["values", "DataFrame"],
        strategy: SearchStrategy[Number],
        dtype: PolarsDataType,
    ) -> None:
        async with redis_cms_async(data) as redis:
            prepared = self._prepare_main_test(
                data,
                redis.key,
                series_names,
                time_zone,
                key_timestamp_value,
                case,
                strategy,
                dtype,
            )
            res_madd = await time_series_madd_async(
                redis.ts,
                prepared.values_or_df,
                key=prepared.key,
                timestamp=prepared.timestamp,
                value=prepared.value,
            )
            assert isinstance(res_madd, list)
            for i in res_madd:
                assert isinstance(i, int)
            res_range = await time_series_range_async(
                redis.ts,
                prepared.keys,
                output_key=prepared.key,
                output_timestamp=prepared.timestamp,
                output_time_zone=time_zone,
                output_value=prepared.value,
            )
            check_polars_dataframe(res_range, height=2, schema_list=prepared.schema)
            assert res_range.rows() == prepared.triples

    @given(data=data())
    async def test_async_error_madd_key_missing(self, *, data: DataObject) -> None:
        df = DataFrame()
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError, match="DataFrame must have a 'key' column; got .*"
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_timestamp_missing(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError,
                match="DataFrame must have a 'timestamp' column; got .*",
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_value_missing(self, *, data: DataObject) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError,
                match="DataFrame must have a 'value' column; got .*",
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_key_is_not_utf8(self, *, data: DataObject) -> None:
        df = DataFrame(
            schema={"key": Boolean, "timestamp": DatetimeUTC, "value": Float64}
        )
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError, match="The 'key' column must be Utf8; got Boolean"
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean, "value": Float64})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError,
                match="The 'timestamp' column must be a zoned Datetime; got Boolean",
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    async def test_async_error_madd_value_is_not_numeric(
        self, *, data: DataObject
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC, "value": Boolean})
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesMAddError,
                match="The 'value' column must be numeric; got Boolean",
            ):
                _ = await time_series_madd_async(redis.ts, df)

    @given(data=data())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    async def test_async_error_madd_invalid_key(
        self, *, data: DataObject, case: Literal["values", "DataFrame"]
    ) -> None:
        async with redis_cms_async(data) as redis:
            values_or_df = self._prepare_test_error_madd_invalid_key(
                data, redis.key, case
            )
            with raises(TimeSeriesMAddError, match="The key '.*' must exist"):
                _ = await time_series_madd_async(
                    redis.ts, values_or_df, assume_time_series_exist=True
                )

    @given(data=data())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    async def test_async_error_madd_invalid_timestamp(
        self, *, data: DataObject, case: Literal["values", "DataFrame"]
    ) -> None:
        async with redis_cms_async(data) as redis:
            values_or_df = self._prepare_test_error_madd_invalid_timestamp(
                data, redis.key, case
            )
            with raises(
                TimeSeriesMAddError,
                match="Timestamps must be at least the Epoch; got .*",
            ):
                _ = await time_series_madd_async(redis.ts, values_or_df)

    @given(data=data())
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    async def test_async_error_madd_invalid_value(
        self, *, data: DataObject, case: Literal["values", "DataFrame"], value: float
    ) -> None:
        async with redis_cms_async(data) as redis:
            values_or_df = self._prepare_test_error_madd_invalid_value(
                data, redis.key, case, value
            )
            with raises(TimeSeriesMAddError, match="The value .* is invalid"):
                _ = await time_series_madd_async(redis.ts, values_or_df)

    @given(data=data())
    async def test_async_error_range_no_keys_requested(
        self, *, data: DataObject
    ) -> None:
        async with redis_cms_async(data) as redis:
            with raises(
                TimeSeriesRangeError, match="At least 1 key must be requested; got .*"
            ):
                _ = await time_series_range_async(redis.ts, [])

    @given(data=data())
    async def test_async_error_range_invalid_key(self, *, data: DataObject) -> None:
        async with redis_cms_async(data) as redis:
            with raises(TimeSeriesRangeError, match="The key '.*' must exist"):
                _ = await time_series_range_async(redis.ts, redis.key)

    @given(data=data())
    async def test_async_error_range_key_with_int64_and_float64(
        self, *, data: DataObject
    ) -> None:
        async with redis_cms_async(data) as redis:
            values = self._prepare_test_error_range_key_with_int64_and_float64(
                data, redis.key
            )
            for vals in values:
                _ = await time_series_madd_async(redis.ts, vals)
            with raises(
                TimeSeriesRangeError,
                match="The key '.*' contains both Int64 and Float64 data",
            ):
                _ = await time_series_range_async(redis.ts, redis.key)

    def _prepare_main_test(
        self,
        data: DataObject,
        redis_key: str,
        series_names: tuple[str, str],
        time_zone: ZoneInfo,
        key_timestamp_value: tuple[str, str, str],
        case: Literal["values", "DataFrame"],
        strategy: SearchStrategy[Number],
        dtype: PolarsDataType,
        /,
    ) -> _TestTimeSeriesMAddAndRangePrepare:
        keys = cast(
            tuple[str, str],
            tuple(f"{redis_key}_{case}_{name}" for name in series_names),
        )
        timestamps = data.draw(tuples(valid_zoned_datetimes, valid_zoned_datetimes))
        values = data.draw(tuples(strategy, strategy))
        triples = list(zip(keys, timestamps, values, strict=True))
        key, timestamp, value = key_timestamp_value
        schema = {
            key: Utf8,
            timestamp: zoned_datetime(time_zone=time_zone),
            value: dtype,
        }
        match case:
            case "values":
                values_or_df = triples
            case "DataFrame":
                values_or_df = DataFrame(triples, schema=schema, orient="row")
        return _TestTimeSeriesMAddAndRangePrepare(
            keys=keys,
            triples=triples,
            key=key,
            timestamp=timestamp,
            value=value,
            values_or_df=values_or_df,
            schema=schema,
        )

    def _prepare_test_error_madd_invalid_key(
        self, data: DataObject, key: str, case: Literal["values", "DataFrame"], /
    ) -> list[tuple[str, dt.datetime, int]] | DataFrame:
        timestamp = data.draw(valid_zoned_datetimes)
        value = data.draw(int32s())
        values = [(f"{key}_{case}", timestamp, value)]
        match case:
            case "values":
                return values
            case "DataFrame":
                return DataFrame(values, schema=self.int_schema, orient="row")

    def _prepare_test_error_madd_invalid_timestamp(
        self, data: DataObject, key: str, case: Literal["values", "DataFrame"], /
    ) -> list[tuple[str, dt.datetime, int]] | DataFrame:
        timestamp = data.draw(invalid_zoned_datetimes)
        _ = assume(timestamp < EPOCH_UTC)
        value = data.draw(int32s())
        values = [(f"{key}_{case}", timestamp, value)]
        match case:
            case "values":
                return values
            case "DataFrame":
                return DataFrame(values, schema=self.int_schema, orient="row")

    def _prepare_test_error_madd_invalid_value(
        self,
        data: DataObject,
        key: str,
        case: Literal["values", "DataFrame"],
        value: float,
        /,
    ) -> list[tuple[str, dt.datetime, float]] | DataFrame:
        timestamp = data.draw(valid_zoned_datetimes)
        values = [(f"{key}_{case}", timestamp, value)]
        match case:
            case "values":
                return values
            case "DataFrame":
                return DataFrame(values, schema=self.float_schema, orient="row")

    def _prepare_test_error_range_key_with_int64_and_float64(
        self, data: DataObject, key: str, /
    ) -> tuple[
        list[tuple[str, dt.datetime, int]], list[tuple[str, dt.datetime, float]]
    ]:
        timestamp = data.draw(valid_zoned_datetimes)
        value = data.draw(int32s())
        return [(key, timestamp, value)], [(key, timestamp, float(value))]


class TestYieldClient:
    def test_sync(self) -> None:
        with yield_client() as client:
            assert isinstance(client, redis.Redis)

    async def test_async(self) -> None:
        async with yield_client_async() as client:
            assert isinstance(client, redis.asyncio.Redis)


class TestYieldTimeSeries:
    def test_sync(self) -> None:
        with yield_time_series() as ts:
            assert isinstance(ts, TimeSeries)

    async def test_async(self) -> None:
        async with yield_time_series_async() as ts:
            assert isinstance(ts, TimeSeries)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from functools import reduce
from itertools import repeat
from typing import TYPE_CHECKING, Any, cast, overload

import numpy as np
from numpy import (
    array,
    bool_,
    digitize,
    dtype,
    errstate,
    exp,
    flatnonzero,
    flip,
    float64,
    full_like,
    inf,
    int64,
    isclose,
    isfinite,
    isinf,
    isnan,
    linspace,
    log,
    nan,
    nanquantile,
    object_,
    prod,
    rint,
    roll,
    where,
)
from numpy.linalg import det, eig
from numpy.random import default_rng
from numpy.typing import NDArray
from typing_extensions import override

from utilities.errors import redirect_error
from utilities.iterables import is_iterable_not_str

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


# RNG
DEFAULT_RNG = default_rng()


# dtypes
datetime64Y = dtype("datetime64[Y]")  # noqa: N816
datetime64M = dtype("datetime64[M]")  # noqa: N816
datetime64W = dtype("datetime64[W]")  # noqa: N816
datetime64D = dtype("datetime64[D]")  # noqa: N816
datetime64h = dtype("datetime64[h]")
datetime64m = dtype("datetime64[m]")
datetime64s = dtype("datetime64[s]")
datetime64ms = dtype("datetime64[ms]")
datetime64us = dtype("datetime64[us]")
datetime64ns = dtype("datetime64[ns]")
datetime64ps = dtype("datetime64[ps]")
datetime64fs = dtype("datetime64[fs]")
datetime64as = dtype("datetime64[as]")


timedelta64Y = dtype("timedelta64[Y]")  # noqa: N816
timedelta64M = dtype("timedelta64[M]")  # noqa: N816
timedelta64W = dtype("timedelta64[W]")  # noqa: N816
timedelta64D = dtype("timedelta64[D]")  # noqa: N816
timedelta64h = dtype("timedelta64[h]")
timedelta64m = dtype("timedelta64[m]")
timedelta64s = dtype("timedelta64[s]")
timedelta64ms = dtype("timedelta64[ms]")
timedelta64us = dtype("timedelta64[us]")
timedelta64ns = dtype("timedelta64[ns]")
timedelta64ps = dtype("timedelta64[ps]")
timedelta64fs = dtype("timedelta64[fs]")
timedelta64as = dtype("timedelta64[as]")


# annotations - dtypes
NDArrayA = NDArray[Any]
NDArrayB = NDArray[bool_]
NDArrayF = NDArray[float64]
NDArrayI = NDArray[int64]
NDArrayO = NDArray[object_]


# functions


def array_indexer(i: int, ndim: int, /, *, axis: int = -1) -> tuple[int | slice, ...]:
    """Get the indexer which returns the `ith` slice of an array along an axis."""
    indexer: list[int | slice] = list(repeat(slice(None), times=ndim))
    indexer[axis] = i
    return tuple(indexer)


def as_int(
    array: NDArrayF, /, *, nan: int | None = None, inf: int | None = None
) -> NDArrayI:
    """Safely cast an array of floats into ints."""
    if (is_nan := isnan(array)).any():
        if nan is None:
            msg = f"{array=}"
            raise AsIntError(msg)
        return as_int(where(is_nan, nan, array).astype(float))
    if (is_inf := isinf(array)).any():
        if inf is None:
            msg = f"{array=}"
            raise AsIntError(msg)
        return as_int(where(is_inf, inf, array).astype(float))
    rounded = rint(array)
    if (isfinite(array) & (~isclose(array, rounded))).any():
        msg = f"{array=}"
        raise AsIntError(msg)
    return rounded.astype(int)


class AsIntError(Exception): ...


def discretize(x: NDArrayF, bins: int | Iterable[float], /) -> NDArrayF:
    """Discretize an array of floats.

    Finite values are mapped to {0, ..., bins-1}.
    """
    if len(x) == 0:
        return array([], dtype=float)
    if isinstance(bins, int):
        bins_use = linspace(0, 1, num=bins + 1)
    else:
        bins_use = array(list(bins), dtype=float)
    if (is_fin := isfinite(x)).all():
        edges = nanquantile(x, bins_use)
        edges[[0, -1]] = [-inf, inf]
        return digitize(x, edges[1:]).astype(float)
    out = full_like(x, nan, dtype=float)
    out[is_fin] = discretize(x[is_fin], bins)
    return out


def ewma(array: NDArrayF, halflife: float, /, *, axis: int = -1) -> NDArrayF:
    """Compute the EWMA of an array."""
    from numbagg import move_exp_nanmean

    alpha = _exp_weighted_alpha(halflife)
    return cast(Any, move_exp_nanmean)(array, axis=axis, alpha=alpha)


def exp_moving_sum(array: NDArrayF, halflife: float, /, *, axis: int = -1) -> NDArrayF:
    """Compute the exponentially-weighted moving sum of an array."""
    from numbagg import move_exp_nansum

    alpha = _exp_weighted_alpha(halflife)
    return cast(Any, move_exp_nansum)(array, axis=axis, alpha=alpha)


def _exp_weighted_alpha(halflife: float, /) -> float:
    """Get the alpha."""
    decay = 1.0 - exp(log(0.5) / halflife)
    com = 1.0 / decay - 1.0
    return 1.0 / (1.0 + com)


def ffill(array: NDArrayF, /, *, limit: int | None = None, axis: int = -1) -> NDArrayF:
    """Forward fill the elements in an array."""
    from bottleneck import push

    return push(array, n=limit, axis=axis)


def ffill_non_nan_slices(
    array: NDArrayF, /, *, limit: int | None = None, axis: int = -1
) -> NDArrayF:
    """Forward fill the slices in an array which contain non-nan values."""
    ndim = array.ndim
    arrays = (
        array[array_indexer(i, ndim, axis=axis)] for i in range(array.shape[axis])
    )
    out = array.copy()
    for i, repl_i in _ffill_non_nan_slices_helper(arrays, limit=limit):
        out[array_indexer(i, ndim, axis=axis)] = repl_i
    return out


def _ffill_non_nan_slices_helper(
    arrays: Iterator[NDArrayF], /, *, limit: int | None = None
) -> Iterator[tuple[int, NDArrayF]]:
    """Yield the slices to be pasted in."""
    last: tuple[int, NDArrayF] | None = None
    for i, arr_i in enumerate(arrays):
        if (~isnan(arr_i)).any():
            last = i, arr_i
        elif last is not None:
            last_i, last_sl = last
            if (limit is None) or ((i - last_i) <= limit):
                yield i, last_sl


def fillna(array: NDArrayF, /, *, value: float = 0.0) -> NDArrayF:
    """Fill the null elements in an array."""
    return where(isnan(array), value, array)


def flatn0(array: NDArrayB, /) -> int:
    """Return the index of the unique True element."""
    if not array.any():
        raise FlatN0EmptyError(array=array)
    flattened = flatnonzero(array)
    try:
        return flattened.item()
    except ValueError:
        raise FlatN0MultipleError(array=array) from None


@dataclass(kw_only=True)
class FlatN0Error(Exception):
    array: NDArrayB


@dataclass(kw_only=True)
class FlatN0EmptyError(FlatN0Error):
    @override
    def __str__(self) -> str:
        return f"Array {self.array} must contain a True."


@dataclass(kw_only=True)
class FlatN0MultipleError(FlatN0Error):
    @override
    def __str__(self) -> str:
        return f"Array {self.array} must contain at most one True."


def has_dtype(x: Any, dtype: Any, /) -> bool:
    """Check if an object has the required dtype."""
    if is_iterable_not_str(dtype):
        return any(has_dtype(x, d) for d in dtype)
    return x.dtype == dtype


def is_at_least(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y."""
    return (x >= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


def is_at_least_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y or x == nan."""
    return is_at_least(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_at_most(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y."""
    return (x <= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


def is_at_most_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y or x == nan."""
    return is_at_most(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_between(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high."""
    return is_at_least(
        x, low, rtol=rtol, atol=atol, equal_nan=equal_nan or low_equal_nan
    ) & is_at_most(x, high, rtol=rtol, atol=atol, equal_nan=equal_nan or high_equal_nan)


def is_between_or_nan(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high or x == nan."""
    return is_between(
        x,
        low,
        high,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
        low_equal_nan=low_equal_nan,
        high_equal_nan=high_equal_nan,
    ) | isnan(x)


def _is_close(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x == y."""
    return np.isclose(
        x,
        y,
        **({} if rtol is None else {"rtol": rtol}),
        **({} if atol is None else {"atol": atol}),
        equal_nan=equal_nan,
    )


def is_empty(shape_or_array: int | tuple[int, ...] | NDArrayA, /) -> bool:
    """Check if an ndarray is empty."""
    if isinstance(shape_or_array, int):
        return shape_or_array == 0
    if isinstance(shape_or_array, tuple):
        return (len(shape_or_array) == 0) or (prod(shape_or_array).item() == 0)
    return is_empty(shape_or_array.shape)


def is_finite_and_integral(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf and x == int(x)."""
    return isfinite(x) & is_integral(x, rtol=rtol, atol=atol)


def is_finite_and_integral_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf and x == int(x), or x == nan."""
    return is_finite_and_integral(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < 0."""
    return isfinite(x) & is_negative(x, rtol=rtol, atol=atol)


def is_finite_and_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < 0 or x == nan."""
    return is_finite_and_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 <= x < inf."""
    return isfinite(x) & is_non_negative(x, rtol=rtol, atol=atol)


def is_finite_and_non_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 <= x < inf or x == nan."""
    return is_finite_and_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x <= 0."""
    return isfinite(x) & is_non_positive(x, rtol=rtol, atol=atol)


def is_finite_and_non_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x <= 0 or x == nan."""
    return is_finite_and_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_zero(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf, x != 0."""
    return isfinite(x) & is_non_zero(x, rtol=rtol, atol=atol)


def is_finite_and_non_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_finite_and_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 < x < inf."""
    return isfinite(x) & is_positive(x, rtol=rtol, atol=atol)


def is_finite_and_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 < x < inf or x == nan."""
    return is_finite_and_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_or_nan(x: Any, /) -> Any:
    """Check if -inf < x < inf or x == nan."""
    return isfinite(x) | isnan(x)


def is_greater_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y."""
    return ((x > y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


def is_greater_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y or x == nan."""
    return is_greater_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_integral(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == int(x)."""
    return _is_close(x, rint(x), rtol=rtol, atol=atol)


def is_integral_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == int(x) or x == nan."""
    return is_integral(x, rtol=rtol, atol=atol) | isnan(x)


def is_less_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y."""
    return ((x < y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


def is_less_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y or x == nan."""
    return is_less_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x < 0."""
    return is_less_than(x, 0.0, rtol=rtol, atol=atol)


def is_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x < 0 or x == nan."""
    return is_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_empty(shape_or_array: int | tuple[int, ...] | NDArrayA, /) -> bool:
    """Check if an ndarray is non-empty."""
    if isinstance(shape_or_array, int):
        return shape_or_array >= 1
    if isinstance(shape_or_array, tuple):
        return (len(shape_or_array) >= 1) and (prod(shape_or_array).item() >= 1)
    return is_non_empty(shape_or_array.shape)


def is_non_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x >= 0."""
    return is_at_least(x, 0.0, rtol=rtol, atol=atol)


def is_non_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x >= 0 or x == nan."""
    return is_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x <= 0."""
    return is_at_most(x, 0.0, rtol=rtol, atol=atol)


def is_non_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x <=0 or x == nan."""
    return is_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_singular(
    array: NDArrayF | NDArrayI,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
) -> bool:
    """Check if det(x) != 0."""
    try:
        with errstate(over="raise"):
            return is_non_zero(det(array), rtol=rtol, atol=atol).item()
    except FloatingPointError:  # pragma: no cover
        return False


def is_non_zero(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0."""
    return ~_is_close(x, 0.0, rtol=rtol, atol=atol)


def is_non_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0."""
    return is_greater_than(x, 0, rtol=rtol, atol=atol)


def is_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_positive_semidefinite(x: NDArrayF | NDArrayI, /) -> bool:
    """Check if `x` is positive semidefinite."""
    if not is_symmetric(x):
        return False
    w, _ = eig(x)
    return bool(is_non_negative(w).all())


def is_symmetric(
    array: NDArrayF | NDArrayI,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> bool:
    """Check if x == x.T."""
    m, n = array.shape
    return (m == n) and (
        _is_close(array, array.T, rtol=rtol, atol=atol, equal_nan=equal_nan)
        .all()
        .item()
    )


def is_zero(x: Any, /, *, rtol: float | None = None, atol: float | None = None) -> Any:
    """Check if x == 0."""
    return _is_close(x, 0.0, rtol=rtol, atol=atol)


def is_zero_or_finite_and_non_micro(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) | is_finite_and_non_zero(x, rtol=rtol, atol=atol)


def is_zero_or_finite_and_non_micro_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0), or x == nan."""
    return is_zero_or_finite_and_non_micro(x, rtol=rtol, atol=atol) | isnan(x)


def is_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_zero_or_non_micro(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0 or ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) | is_non_zero(x, rtol=rtol, atol=atol)


def is_zero_or_non_micro_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0 or ~isclose(x, 0) or x == nan."""
    return is_zero_or_non_micro(x, rtol=rtol, atol=atol) | isnan(x)


@overload
def maximum(x: float, /) -> float: ...
@overload
def maximum(x0: float, x1: float, /) -> float: ...
@overload
def maximum(x0: float, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: float, x2: float, /) -> float: ...
@overload
def maximum(x0: float, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
def maximum(*xs: float | NDArrayF) -> float | NDArrayF:
    """Compute the maximum of a number of quantities."""
    return reduce(np.maximum, xs)


@overload
def minimum(x: float, /) -> float: ...
@overload
def minimum(x0: float, x1: float, /) -> float: ...
@overload
def minimum(x0: float, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: float, x2: float, /) -> float: ...
@overload
def minimum(x0: float, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
def minimum(*xs: float | NDArrayF) -> float | NDArrayF:
    """Compute the minimum of a number of quantities."""
    return reduce(np.minimum, xs)


def pct_change(
    array: NDArrayF | NDArrayI,
    /,
    *,
    limit: int | None = None,
    n: int = 1,
    axis: int = -1,
) -> NDArrayF:
    """Compute the percentage change in an array."""
    if n == 0:
        raise PctChangeError
    if n > 0:
        filled = ffill(array.astype(float), limit=limit, axis=axis)
        shifted = shift(filled, n=n, axis=axis)
        with errstate(all="ignore"):
            ratio = (filled / shifted) if n >= 0 else (shifted / filled)
        return where(isfinite(array), ratio - 1.0, nan)
    flipped = cast(NDArrayF | NDArrayI, flip(array, axis=axis))
    result = pct_change(flipped, limit=limit, n=-n, axis=axis)
    return flip(result, axis=axis)


@dataclass(kw_only=True)
class PctChangeError(Exception):
    @override
    def __str__(self) -> str:
        return "Shift must be non-zero"


@contextmanager
def redirect_empty_numpy_concatenate() -> Iterator[None]:
    """Redirect to the `EmptyNumpyConcatenateError`."""
    with redirect_error(
        ValueError,
        EmptyNumpyConcatenateError,
        match="need at least one array to concatenate",
    ):
        yield


class EmptyNumpyConcatenateError(Exception): ...


def shift(array: NDArrayF | NDArrayI, /, *, n: int = 1, axis: int = -1) -> NDArrayF:
    """Shift the elements of an array."""
    if n == 0:
        raise ShiftError
    as_float = array.astype(float)
    shifted = roll(as_float, n, axis=axis)
    indexer = list(repeat(slice(None), times=array.ndim))
    indexer[axis] = slice(n) if n >= 0 else slice(n, None)
    shifted[tuple(indexer)] = nan
    return shifted


@dataclass(kw_only=True)
class ShiftError(Exception):
    @override
    def __str__(self) -> str:
        return "Shift must be non-zero"


def shift_bool(
    array: NDArrayB, /, *, n: int = 1, axis: int = -1, fill_value: bool = False
) -> NDArrayB:
    """Shift the elements of a boolean array."""
    shifted = shift(array.astype(float), n=n, axis=axis)
    return fillna(shifted, value=float(fill_value)).astype(bool)


__all__ = [
    "DEFAULT_RNG",
    "AsIntError",
    "EmptyNumpyConcatenateError",
    "FlatN0EmptyError",
    "FlatN0Error",
    "FlatN0MultipleError",
    "NDArrayA",
    "NDArrayB",
    "NDArrayF",
    "NDArrayI",
    "NDArrayO",
    "PctChangeError",
    "ShiftError",
    "array_indexer",
    "as_int",
    "datetime64D",
    "datetime64M",
    "datetime64W",
    "datetime64Y",
    "datetime64as",
    "datetime64fs",
    "datetime64h",
    "datetime64m",
    "datetime64ms",
    "datetime64ns",
    "datetime64ps",
    "datetime64s",
    "datetime64us",
    "discretize",
    "ewma",
    "exp_moving_sum",
    "ffill",
    "ffill_non_nan_slices",
    "fillna",
    "flatn0",
    "has_dtype",
    "is_at_least",
    "is_at_least_or_nan",
    "is_at_most",
    "is_at_most_or_nan",
    "is_between",
    "is_between_or_nan",
    "is_empty",
    "is_finite_and_integral",
    "is_finite_and_integral_or_nan",
    "is_finite_and_negative",
    "is_finite_and_negative_or_nan",
    "is_finite_and_non_negative",
    "is_finite_and_non_negative_or_nan",
    "is_finite_and_non_positive",
    "is_finite_and_non_positive_or_nan",
    "is_finite_and_non_zero",
    "is_finite_and_non_zero_or_nan",
    "is_finite_and_positive",
    "is_finite_and_positive_or_nan",
    "is_finite_or_nan",
    "is_greater_than",
    "is_greater_than_or_nan",
    "is_integral",
    "is_integral_or_nan",
    "is_less_than",
    "is_less_than_or_nan",
    "is_negative",
    "is_negative_or_nan",
    "is_non_empty",
    "is_non_negative",
    "is_non_negative_or_nan",
    "is_non_positive",
    "is_non_positive_or_nan",
    "is_non_singular",
    "is_non_zero",
    "is_non_zero_or_nan",
    "is_positive",
    "is_positive_or_nan",
    "is_positive_semidefinite",
    "is_symmetric",
    "is_zero",
    "is_zero_or_finite_and_non_micro",
    "is_zero_or_finite_and_non_micro_or_nan",
    "is_zero_or_nan",
    "is_zero_or_non_micro",
    "is_zero_or_non_micro_or_nan",
    "maximum",
    "minimum",
    "pct_change",
    "redirect_empty_numpy_concatenate",
    "shift",
    "shift_bool",
]



from __future__ import annotations

import builtins
import datetime as dt
from collections.abc import (
    AsyncIterator,
    Callable,
    Collection,
    Hashable,
    Iterable,
    Iterator,
)
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    asynccontextmanager,
    contextmanager,
    suppress,
)
from dataclasses import dataclass
from datetime import timezone
from enum import Enum, auto
from math import ceil, floor, inf, isfinite, nan
from os import environ
from pathlib import Path
from re import search
from string import ascii_letters, printable
from subprocess import run
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
    assert_never,
    cast,
    overload,
)
from zoneinfo import ZoneInfo

import redis
import redis.asyncio
from hypothesis import HealthCheck, Phase, Verbosity, assume, settings
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import (
    DataObject,
    DrawFn,
    SearchStrategy,
    booleans,
    characters,
    composite,
    dates,
    datetimes,
    floats,
    integers,
    just,
    lists,
    none,
    sampled_from,
    text,
    timedeltas,
    uuids,
)
from hypothesis.utils.conventions import not_set
from redis.exceptions import ResponseError
from redis.typing import KeyT

from utilities.datetime import MAX_MONTH, MIN_MONTH, Month, date_to_month, get_now
from utilities.math import MAX_INT32, MAX_INT64, MAX_UINT32, MIN_INT32, MIN_INT64
from utilities.pathlib import temp_cwd
from utilities.platform import IS_WINDOWS
from utilities.tempfile import TEMP_DIR, TemporaryDirectory
from utilities.text import ensure_str
from utilities.whenever import CheckValidZonedDateimeError, check_valid_zoned_datetime
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from uuid import UUID

    from hypothesis.database import ExampleDatabase
    from numpy.random import RandomState
    from redis.commands.timeseries import TimeSeries
    from sqlalchemy import Engine, MetaData
    from sqlalchemy.ext.asyncio import AsyncEngine

    from utilities.numpy import NDArrayB, NDArrayF, NDArrayI, NDArrayO
    from utilities.types import Duration, Number


_T = TypeVar("_T")
MaybeSearchStrategy = _T | SearchStrategy[_T]
Shape = int | tuple[int, ...]


async def aiosqlite_engines(
    data: DataObject, /, *, metadata: MetaData | None = None, base: Any = None
) -> AsyncEngine:
    from utilities.sqlalchemy import create_engine

    temp_path = data.draw(temp_paths())
    path = Path(temp_path, "db.sqlite")
    engine = create_engine("sqlite+aiosqlite", database=str(path), async_=True)
    if metadata is not None:
        async with engine.begin() as conn:  # pragma: no cover
            await conn.run_sync(metadata.create_all)
    if base is not None:
        async with engine.begin() as conn:  # pragma: no cover
            await conn.run_sync(base.metadata.create_all)

    class EngineWithPath(type(engine)): ...

    engine_with_path = EngineWithPath(engine.sync_engine)
    cast(Any, engine_with_path).temp_path = temp_path  # keep `temp_path` alive
    return engine_with_path


@contextmanager
def assume_does_not_raise(
    *exceptions: type[Exception], match: str | None = None
) -> Iterator[None]:
    """Assume a set of exceptions are not raised.

    Optionally filter on the string representation of the exception.
    """
    try:
        yield
    except exceptions as caught:
        if match is None:
            _ = assume(condition=False)
        else:
            (msg,) = caught.args
            if search(match, ensure_str(msg)):
                _ = assume(condition=False)
            else:
                raise


@composite
def bool_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] | None = None,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayB:
    """Strategy for generating arrays of booleans."""
    from hypothesis.extra.numpy import array_shapes, arrays

    draw = lift_draw(_draw)
    shape_use = array_shapes() if shape is None else shape
    strategy: SearchStrategy[NDArrayB] = arrays(
        bool, draw(shape_use), elements=booleans(), fill=fill, unique=draw(unique)
    )
    return draw(strategy)


@composite
def durations(
    _draw: DrawFn,
    /,
    *,
    min_number: MaybeSearchStrategy[Number] | None = None,
    max_number: MaybeSearchStrategy[Number] | None = None,
    min_timedelta: MaybeSearchStrategy[dt.timedelta] = dt.timedelta.min,
    max_timedelta: MaybeSearchStrategy[dt.timedelta] = dt.timedelta.max,
    two_way: bool = False,
) -> Duration:
    """Strategy for generating durations."""
    draw = lift_draw(_draw)
    min_number_, max_number_ = draw(min_number), draw(max_number)
    min_timedelta_, max_timedelta_ = draw(min_timedelta), draw(max_timedelta)
    st_integers = integers(
        min_value=min_number_ if isinstance(min_number_, int) else None,
        max_value=max_number_ if isinstance(max_number_, int) else None,
    )
    st_floats = floats(
        min_value=min_number_,
        max_value=max_number_,
        allow_nan=False,
        allow_infinity=False,
    )
    if two_way:
        from utilities.whenever import MAX_TWO_WAY_TIMEDELTA, MIN_TWO_WAY_TIMEDELTA

        min_timedelta_ = max(min_timedelta_, MIN_TWO_WAY_TIMEDELTA)
        max_timedelta_ = min(max_timedelta_, MAX_TWO_WAY_TIMEDELTA)
    st_timedeltas = timedeltas(min_value=min_timedelta_, max_value=max_timedelta_)
    return _draw(st_integers | st_floats | st_timedeltas)


@composite
def float_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] | None = None,
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayF:
    """Strategy for generating arrays of floats."""
    from hypothesis.extra.numpy import array_shapes, arrays

    draw = lift_draw(_draw)
    shape_use = array_shapes() if shape is None else shape
    elements = floats_extra(
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_inf=allow_inf,
        allow_pos_inf=allow_pos_inf,
        allow_neg_inf=allow_neg_inf,
        integral=integral,
    )
    strategy: SearchStrategy[NDArrayF] = arrays(
        float, draw(shape_use), elements=elements, fill=fill, unique=draw(unique)
    )
    return draw(strategy)


@composite
def floats_extra(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
) -> float:
    """Strategy for generating floats, with extra special values."""
    draw = lift_draw(_draw)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    elements = floats(
        min_value=min_value_,
        max_value=max_value_,
        allow_nan=False,
        allow_infinity=False,
    )
    if draw(allow_nan):
        elements |= just(nan)
    if draw(allow_inf):
        elements |= sampled_from([inf, -inf])
    if draw(allow_pos_inf):
        elements |= just(inf)
    if draw(allow_neg_inf):
        elements |= just(-inf)
    element = draw(elements)
    if isfinite(element) and draw(integral):
        candidates = [floor(element), ceil(element)]
        if min_value_ is not None:
            candidates = [c for c in candidates if c >= min_value_]
        if max_value_ is not None:
            candidates = [c for c in candidates if c <= max_value_]
        _ = assume(len(candidates) >= 1)
        element = draw(sampled_from(candidates))
        return float(element)
    return element


@composite
def git_repos(
    _draw: DrawFn, /, *, branch: MaybeSearchStrategy[str | None] = None
) -> Path:
    draw = lift_draw(_draw)
    path = draw(temp_paths())
    with temp_cwd(path):
        _ = run(["git", "init"], check=True)
        _ = run(["git", "config", "user.name", "User"], check=True)
        _ = run(["git", "config", "user.email", "a@z.com"], check=True)
        file = Path(path, "file")
        file.touch()
        file_str = str(file)
        _ = run(["git", "add", file_str], check=True)
        _ = run(["git", "commit", "-m", "add"], check=True)
        _ = run(["git", "rm", file_str], check=True)
        _ = run(["git", "commit", "-m", "rm"], check=True)
        if (branch := draw(branch)) is not None:
            _ = run(["git", "checkout", "-b", branch], check=True)
    return path


def hashables() -> SearchStrategy[Hashable]:
    """Strategy for generating hashable elements."""
    return booleans() | integers() | none() | text_ascii()


@composite
def int_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] | None = None,
    min_value: MaybeSearchStrategy[int] = MIN_INT64,
    max_value: MaybeSearchStrategy[int] = MAX_INT64,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayI:
    """Strategy for generating arrays of ints."""
    from hypothesis.extra.numpy import array_shapes, arrays
    from numpy import int64

    draw = lift_draw(_draw)
    shape_use = array_shapes() if shape is None else shape
    elements = int64s(min_value=min_value, max_value=max_value)
    strategy: SearchStrategy[NDArrayI] = arrays(
        int64, draw(shape_use), elements=elements, fill=fill, unique=draw(unique)
    )
    return draw(strategy)


@composite
def int32s(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[int] = MIN_INT32,
    max_value: MaybeSearchStrategy[int] = MAX_INT32,
) -> int:
    """Strategy for generating int32s."""
    draw = lift_draw(_draw)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_ = max(min_value_, MIN_INT32)
    max_value_ = min(max_value_, MAX_INT32)
    return draw(integers(min_value_, max_value_))


@composite
def int64s(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[int] = MIN_INT64,
    max_value: MaybeSearchStrategy[int] = MAX_INT64,
) -> int:
    """Strategy for generating int64s."""
    draw = lift_draw(_draw)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_ = max(min_value_, MIN_INT64)
    max_value_ = min(max_value_, MAX_INT64)
    return draw(integers(min_value_, max_value_))


_MDF = TypeVar("_MDF")


class _MaybeDrawFn(Protocol):
    @overload
    def __call__(self, obj: SearchStrategy[_MDF], /) -> _MDF: ...
    @overload
    def __call__(self, obj: MaybeSearchStrategy[_MDF], /) -> _MDF: ...
    def __call__(self, obj: MaybeSearchStrategy[_MDF], /) -> _MDF:
        raise NotImplementedError(obj)  # pragma: no cover


def lift_draw(draw: DrawFn, /) -> _MaybeDrawFn:
    """Lift the `draw` function to handle non-`SearchStrategy` types."""

    def func(obj: MaybeSearchStrategy[_MDF], /) -> _MDF:
        return draw(obj) if isinstance(obj, SearchStrategy) else obj

    return cast(Any, func)


@composite
def lists_fixed_length(
    _draw: DrawFn,
    strategy: SearchStrategy[_T],
    size: MaybeSearchStrategy[int],
    /,
    *,
    unique: MaybeSearchStrategy[bool] = False,
    sorted: MaybeSearchStrategy[bool] = False,  # noqa: A002
) -> list[_T]:
    """Strategy for generating lists of a fixed length."""
    draw = lift_draw(_draw)
    size_ = draw(size)
    elements = draw(
        lists(strategy, min_size=size_, max_size=size_, unique=draw(unique))
    )
    if draw(sorted):
        return builtins.sorted(cast(Iterable[Any], elements))
    return elements


@composite
def months(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[Month] = MIN_MONTH,
    max_value: MaybeSearchStrategy[Month] = MAX_MONTH,
) -> Month:
    """Strategy for generating datetimes with the UTC timezone."""
    draw = lift_draw(_draw)
    min_date = draw(min_value).to_date()
    max_date = draw(max_value).to_date()
    date = draw(dates(min_value=min_date, max_value=max_date))
    return date_to_month(date)


@composite
def random_states(
    _draw: DrawFn, /, *, seed: MaybeSearchStrategy[int | None] = None
) -> RandomState:
    """Strategy for generating `numpy` random states."""
    from numpy.random import RandomState

    draw = lift_draw(_draw)
    seed_ = draw(seed)
    seed_use = draw(integers(0, MAX_UINT32)) if seed_ is None else seed_
    return RandomState(seed=seed_use)


_TRedis = TypeVar("_TRedis", redis.Redis, redis.asyncio.Redis)


@dataclass(repr=False, frozen=True, kw_only=True)
class RedisContainer(Generic[_TRedis]):
    """A container for `redis.Client`."""

    client: _TRedis
    timestamp: dt.datetime
    uuid: UUID
    key: str

    @property
    def ts(self) -> TimeSeries:
        return self.client.ts()  # skipif-ci-and-not-linux


YieldRedisContainer = Callable[[], AbstractContextManager[RedisContainer[redis.Redis]]]


@composite
def redis_cms(draw: DrawFn, /) -> YieldRedisContainer:
    """Strategy for generating redis clients (with cleanup)."""
    from redis import Redis  # skipif-ci-and-not-linux
    from redis.exceptions import ResponseError  # skipif-ci-and-not-linux

    now = get_now(time_zone="local")  # skipif-ci-and-not-linux
    uuid = draw(uuids())  # skipif-ci-and-not-linux
    key = f"{now}_{uuid}"  # skipif-ci-and-not-linux

    @contextmanager
    def yield_redis() -> (  # skipif-ci-and-not-linux
        Iterator[RedisContainer[redis.Redis]]
    ):
        with Redis(db=15) as client:  # skipif-ci-and-not-linux
            keys = cast(list[KeyT], client.keys(pattern=f"{key}_*"))
            with suppress(ResponseError):
                _ = client.delete(*keys)
            yield RedisContainer(client=client, timestamp=now, uuid=uuid, key=key)
            keys = cast(list[KeyT], client.keys(pattern=f"{key}_*"))
            with suppress(ResponseError):
                _ = client.delete(*keys)

    return yield_redis  # skipif-ci-and-not-linux


def redis_cms_async(
    data: DataObject, /
) -> AbstractAsyncContextManager[RedisContainer[redis.asyncio.Redis]]:
    """Strategy for generating asynchronous redis clients."""
    from redis.asyncio import Redis  # skipif-ci-and-not-linux

    now = get_now(time_zone="local")  # skipif-ci-and-not-linux
    uuid = data.draw(uuids())  # skipif-ci-and-not-linux
    key = f"{now}_{uuid}"  # skipif-ci-and-not-linux

    @asynccontextmanager
    async def yield_redis_async() -> (  # skipif-ci-and-not-linux
        AsyncIterator[RedisContainer[redis.asyncio.Redis]]
    ):
        async with Redis(db=15) as client:  # skipif-ci-and-not-linux
            keys = cast(list[KeyT], await client.keys(pattern=f"{key}_*"))
            with suppress(ResponseError):
                _ = await client.delete(*keys)
            yield RedisContainer(client=client, timestamp=now, uuid=uuid, key=key)
            keys = cast(list[KeyT], await client.keys(pattern=f"{key}_*"))
            with suppress(ResponseError):
                _ = await client.delete(*keys)

    return yield_redis_async()  # skipif-ci-and-not-linux


def setup_hypothesis_profiles(
    *, suppress_health_check: Iterable[HealthCheck] = ()
) -> None:
    """Set up the hypothesis profiles."""

    class Profile(Enum):
        dev = auto()
        default = auto()
        ci = auto()
        debug = auto()

        @property
        def max_examples(self) -> int:
            match self:
                case Profile.dev | Profile.debug:
                    return 10
                case Profile.default:
                    return 100
                case Profile.ci:
                    return 1000
                case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
                    assert_never(never)

        @property
        def verbosity(self) -> Verbosity | None:
            match self:
                case Profile.dev | Profile.default | Profile.ci:
                    return Verbosity.normal
                case Profile.debug:
                    return Verbosity.debug
                case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
                    assert_never(never)

    phases = {Phase.explicit, Phase.reuse, Phase.generate, Phase.target}
    if "HYPOTHESIS_NO_SHRINK" not in environ:
        phases.add(Phase.shrink)
    for profile in Profile:
        try:
            max_examples = int(environ["HYPOTHESIS_MAX_EXAMPLES"])
        except KeyError:
            max_examples = profile.max_examples
        settings.register_profile(
            profile.name,
            max_examples=max_examples,
            phases=phases,
            report_multiple_bugs=True,
            deadline=None,
            print_blob=True,
            suppress_health_check=suppress_health_check,
            verbosity=profile.verbosity,
        )
    settings.load_profile(Profile.default.name)


def settings_with_reduced_examples(
    frac: float = 0.1,
    /,
    *,
    derandomize: bool = not_set,  # pyright: ignore[reportArgumentType]
    database: ExampleDatabase | None = not_set,  # pyright: ignore[reportArgumentType]
    verbosity: Verbosity = not_set,  # pyright: ignore[reportArgumentType]
    phases: Collection[Phase] = not_set,  # pyright: ignore[reportArgumentType]
    stateful_step_count: int = not_set,  # pyright: ignore[reportArgumentType]
    report_multiple_bugs: bool = not_set,  # pyright: ignore[reportArgumentType]
    suppress_health_check: Collection[HealthCheck] = not_set,  # pyright: ignore[reportArgumentType]
    deadline: float | dt.timedelta | None = not_set,  # pyright: ignore[reportArgumentType]
    print_blob: bool = not_set,  # pyright: ignore[reportArgumentType]
    backend: str = not_set,  # pyright: ignore[reportArgumentType]
) -> settings:
    """Set a test to fewer max examples."""
    curr = settings()
    max_examples = max(round(frac * curr.max_examples), 1)
    return settings(
        max_examples=max_examples,
        derandomize=derandomize,
        database=database,
        verbosity=verbosity,
        phases=phases,
        stateful_step_count=stateful_step_count,
        report_multiple_bugs=report_multiple_bugs,
        suppress_health_check=suppress_health_check,
        deadline=deadline,
        print_blob=print_blob,
        backend=backend,
    )


@composite
def slices(
    _draw: DrawFn,
    iter_len: MaybeSearchStrategy[int],
    /,
    *,
    slice_len: MaybeSearchStrategy[int | None] = None,
) -> slice:
    """Strategy for generating continuous slices from an iterable."""
    draw = lift_draw(_draw)
    iter_len_ = draw(iter_len)
    if (slice_len_ := draw(slice_len)) is None:
        slice_len_ = draw(integers(0, iter_len_))
    elif not 0 <= slice_len_ <= iter_len_:
        msg = f"Slice length {slice_len_} exceeds iterable length {iter_len_}"
        raise InvalidArgument(msg)
    start = draw(integers(0, iter_len_ - slice_len_))
    stop = start + slice_len_
    return slice(start, stop)


@composite
def sqlite_engines(
    _draw: DrawFn, /, *, metadata: MetaData | None = None, base: Any = None
) -> Engine:
    """Strategy for generating SQLite engines."""
    from utilities.sqlalchemy import create_engine

    temp_path = _draw(temp_paths())
    path = Path(temp_path, "db.sqlite")
    engine = create_engine("sqlite", database=str(path))
    cast(Any, engine).temp_path = temp_path  # keep `temp_path` alive
    if metadata is not None:
        metadata.create_all(engine)
    if base is not None:
        base.metadata.create_all(engine)
    return engine


@composite
def str_arrays(
    _draw: DrawFn,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] | None = None,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    allow_none: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArrayO:
    """Strategy for generating arrays of strings."""
    from hypothesis.extra.numpy import array_shapes, arrays

    draw = lift_draw(_draw)
    shape_use = array_shapes() if shape is None else shape
    elements = text_ascii(min_size=min_size, max_size=max_size)
    if draw(allow_none):
        elements |= none()
    strategy: SearchStrategy[NDArrayO] = arrays(
        object, draw(shape_use), elements=elements, fill=fill, unique=draw(unique)
    )
    return draw(strategy)


_TEMP_DIR_HYPOTHESIS = Path(TEMP_DIR, "hypothesis")


@composite
def temp_dirs(_draw: DrawFn, /) -> TemporaryDirectory:
    """Search strategy for temporary directories."""
    _TEMP_DIR_HYPOTHESIS.mkdir(exist_ok=True)
    uuid = _draw(uuids())
    return TemporaryDirectory(
        prefix=f"{uuid}__", dir=_TEMP_DIR_HYPOTHESIS, ignore_cleanup_errors=IS_WINDOWS
    )


@composite
def temp_paths(_draw: DrawFn, /) -> Path:
    """Search strategy for paths to temporary directories."""
    temp_dir = _draw(temp_dirs())
    root = temp_dir.path
    cls = type(root)

    class SubPath(cls):
        _temp_dir = temp_dir

    return SubPath(root)


def text_ascii(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[str]:
    """Strategy for generating ASCII text."""
    return _draw_text(
        characters(whitelist_categories=[], whitelist_characters=ascii_letters),
        min_size=min_size,
        max_size=max_size,
        disallow_na=disallow_na,
    )


def text_clean(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[str]:
    """Strategy for generating clean text."""
    return _draw_text(
        characters(blacklist_categories=["Z", "C"]),
        min_size=min_size,
        max_size=max_size,
        disallow_na=disallow_na,
    )


def text_printable(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[str]:
    """Strategy for generating printable text."""
    return _draw_text(
        characters(whitelist_categories=[], whitelist_characters=printable),
        min_size=min_size,
        max_size=max_size,
        disallow_na=disallow_na,
    )


@composite
def timedeltas_2w(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.timedelta] = dt.timedelta.min,
    max_value: MaybeSearchStrategy[dt.timedelta] = dt.timedelta.max,
) -> dt.timedelta:
    """Strategy for generating timedeltas which can be se/deserialized."""
    from utilities.whenever import MAX_TWO_WAY_TIMEDELTA, MIN_TWO_WAY_TIMEDELTA

    draw = lift_draw(_draw)
    min_value_ = max(draw(min_value), MIN_TWO_WAY_TIMEDELTA)
    max_value_ = min(draw(max_value), MAX_TWO_WAY_TIMEDELTA)
    return draw(timedeltas(min_value=min_value_, max_value=max_value_))


_ZONED_DATETIMES_LEFT_MOST = ZoneInfo("Asia/Manila")
_ZONED_DATETIMES_RIGHT_MOST = ZoneInfo("Pacific/Kiritimati")


@composite
def zoned_datetimes(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = dt.datetime.min,
    max_value: MaybeSearchStrategy[dt.datetime] = dt.datetime.max,
    time_zone: MaybeSearchStrategy[ZoneInfo | timezone] = UTC,
    valid: bool = False,
) -> dt.datetime:
    """Strategy for generating zoned datetimes."""
    draw = lift_draw(_draw)
    min_value_, max_value_, time_zone_ = (
        draw(min_value),
        draw(max_value),
        draw(time_zone),
    )
    if min_value_.tzinfo is not None:
        with assume_does_not_raise(OverflowError, match="date value out of range"):
            min_value_ = min_value_.astimezone(time_zone_)
        min_value_ = min_value_.replace(tzinfo=None)
    if max_value_.tzinfo is not None:
        with assume_does_not_raise(OverflowError, match="date value out of range"):
            max_value_ = max_value_.astimezone(time_zone_)
        max_value_ = max_value_.replace(tzinfo=None)
    strategy = datetimes(
        min_value=min_value_, max_value=max_value_, timezones=just(time_zone_)
    )
    datetime = draw(strategy)
    with assume_does_not_raise(OverflowError, match="date value out of range"):
        _ = datetime.astimezone(_ZONED_DATETIMES_LEFT_MOST)  # for dt.datetime.min
    with assume_does_not_raise(OverflowError, match="date value out of range"):
        _ = datetime.astimezone(  # for dt.datetime.max
            _ZONED_DATETIMES_RIGHT_MOST
        )
    result = datetime.astimezone(time_zone_)
    if valid:
        with assume_does_not_raise(  # skipif-ci-and-windows
            CheckValidZonedDateimeError
        ):
            check_valid_zoned_datetime(result)
    return result


@composite
def _draw_text(
    _draw: DrawFn,
    alphabet: MaybeSearchStrategy[str],
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> str:
    """Draw from a text-generating strategy."""
    draw = lift_draw(_draw)
    drawn = draw(text(alphabet, min_size=draw(min_size), max_size=draw(max_size)))
    if draw(disallow_na):
        _ = assume(drawn != "NA")
    return drawn


__all__ = [
    "MaybeSearchStrategy",
    "RedisContainer",
    "Shape",
    "aiosqlite_engines",
    "assume_does_not_raise",
    "bool_arrays",
    "durations",
    "float_arrays",
    "floats_extra",
    "git_repos",
    "hashables",
    "int32s",
    "int64s",
    "int_arrays",
    "lift_draw",
    "lists_fixed_length",
    "months",
    "random_states",
    "redis_cms",
    "redis_cms_async",
    "setup_hypothesis_profiles",
    "slices",
    "sqlite_engines",
    "str_arrays",
    "temp_dirs",
    "temp_paths",
    "text_ascii",
    "text_clean",
    "text_printable",
    "timedeltas_2w",
    "zoned_datetimes",
]



from __future__ import annotations

from typing import Any

from pytest import mark, param

from utilities.functions import identity
from utilities.inspect import bind_args_custom_repr


class TestBindArgsCustomRepr:
    @mark.parametrize(
        ("obj", "expected"),
        [
            param(None, "obj=None"),
            param([], "obj=[]"),
            param([1], "obj=[1]"),
            param([1, 2], "obj=[1, 2]"),
            param([1, 2, 3], "obj=[1, 2, 3]"),
            param([1, 2, 3, 4], "obj=[1, 2, 3, 4]"),
            param([1, 2, 3, 4, 5], "obj=[1, 2, 3, 4, 5]"),
            param([1, 2, 3, 4, 5, 6], "obj=[1, 2, 3, 4, 5, 6]"),
            param([1, 2, 3, 4, 5, 6, 7], "obj=[1, 2, 3, 4, 5, 6, ...]"),
            param([1, 2, 3, 4, 5, 6, 7, 8], "obj=[1, 2, 3, 4, 5, 6, ...]"),
        ],
    )
    def test_main(self, *, obj: Any, expected: str) -> None:
        result = bind_args_custom_repr(identity, obj)
        assert result == expected
